// Copilot main
//
// Copyright (c) 1996 Greg Hewgill

#include <windows.h>
#pragma hdrstop

#include "lcd.h"
#include "pilotcpu\pilotcpu.h"
#include "settings.h"

#include "resource.h"

static const char *CpuInitErrors[] = {
  "",
  "Pilot.ROM file not found",
  "Error loading Pilot.ROM file",
  "Error loading Pilot.RAM file",
};

enum RomTypes  { Blank, Garbage, Pilot, Palm_Bank1, Palm_Bank2, Unknown };
char *RomTypeNames[] = { "Blank", 
						 "Truncated", 
						 "Pilot 1.0 512kB ROM",
						 "PalmPilot 2.0 1MB ROM (First Half)",
						 "PalmPilot 2.0 1MB ROM (Second Half)",
						 "Unknown 512kB image" };

const char *pszImpossible = "!Impossible combination, Copilot internal error";

const char *RomMistakes[6][6] = { 
	{ "!Your ROM image is an empty file.  You might have made a mistake in the concatenation of two Palm ROM halves.",
		pszImpossible,
		pszImpossible,
		pszImpossible,
		pszImpossible,
		pszImpossible 
	},
	{ "!Your ROM image is truncated.  The most likely cause is that you did not use the /B option to copy when concatenating the two Palm ROM halves.",
	    pszImpossible,
	    pszImpossible,
	    pszImpossible,
	    pszImpossible,
	    pszImpossible
	},
	{ ".You have a valid Pilot 1.0 ROM image.",
		"!Your ROM image might work as a Pilot 1.0 ROM, but it seems to have some unexpected data following the 512kB ROM image.",
		"!Your ROM image is a concatenation of two Pilot 1.0 ROMs.  This may run as Pilot 1.0, but is probably not what you meant to do.",
		"!Your ROM image is a concatenation of a Pilot 1.0 ROM and a PalmPilot ROM.  This may run as Pilot 1.0, but is probably not what you meant to do.",
		"!Your ROM image is a concatenation of a Pilot 1.0 ROM and a PalmPilot ROM.  This may run as Pilot 1.0, but is probably not what you meant to do.  You probably didn't get a new ROM for the first half of the image, and reused your old Pilot ROM image one, isntead.",
		"!Your ROM image might work as a Pilot 1.0 ROM, but it seems to have an additional, unexpected image following it."
	},
	{ "!Your ROM image is only the first half of a PalmPilot 2.0 ROM",
	    "!The second half of your ROM images is truncated, but the first half is part of a PalmPilot 2.0 ROM.",
	    "!Your ROM image is a concatenation of a PalmPilot ROM and a Pilot 1.0 ROM.  You may have accidentally reused your old Pilot ROM image.",
	    "!Your ROM image is a concatenation of two first halves of a  PalmPilot ROM.",
	    ".Your ROM image is a valid PalmPilot 2.0 ROM.",
    	"!Your ROM image might work as a PalmPilot 2.0 ROM, but it seems to have an unrecognized second half."
	},
	{ "!Your ROM image is only the second half of a PalmPilot 2.0 ROM",
		"!Your ROM image appears to be a concatenation of a second half with truncated garbage.",
		"!Your ROM image appears to be a concatenation of a PalmPilot 2.0 second half with a Pilot 1.0 image.",
		"!Your ROM image appears to be a reversed concatenation of the two PalmPilot 2.0 halves.",
	    "!Your ROM image is a concatenation of two second halves of a  PalmPilot ROM.",
		"!Your ROM image appears to be a concatenation of a second half with garbage."
	},
	{
		"!Your ROM image is unrecognized by Copilot.  If this is a ROM from a new PalmOS, you should Continue.",
		"!Your ROM image is unrecognized by Copilot.  If this is a ROM from a new PalmOS, you should Continue.",
		"!Your ROM image is unrecognized by Copilot.  If this is a ROM from a new PalmOS, you should Continue.",
		"!Your ROM image is unrecognized by Copilot.  If this is a ROM from a new PalmOS, you should Continue.",
		"!Your ROM image is unrecognized by Copilot.  If this is a ROM from a new PalmOS, you should Continue.",
		"!Your ROM image is unrecognized by Copilot.  If this is a ROM from a new PalmOS, you should Continue."
	}
};








RomTypes ROMs[2] = { Blank, Blank };

static bool load_rom( char * name, UWORD **promemory, unsigned *promsize);

enum RomTypes GetRomType( char *pRom )
{
	if ( 0 == strcmp( pRom + 0x8CE , "17:05:28" ))
		return( Palm_Bank1 );

	if ( 0 == strcmp( pRom + 0x6252E, "Expense" ))
		return( Palm_Bank2 );

	if ( 0 == strcmp( pRom +  0x89E , "10:38:02" ))
		return( Pilot );

	return( Unknown );
}

BOOL CALLBACK RomDlgProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch (msg) {
    case WM_COMMAND:
		switch (LOWORD(wparam)) {
        case IDOK:
        case IDCANCEL:
			EndDialog(hwnd, LOWORD(wparam));
			break;
		}
		break;
    case WM_INITDIALOG:
		{
			HWND hwndDesktop = GetDesktopWindow();

			if ( NULL != hwndDesktop )
			{
				int x, y;
				RECT rc;
				GetWindowRect( hwndDesktop, &rc );
				x = ( rc.right - rc.left )/2;
				y = ( rc.bottom - rc.top )/2;
				GetWindowRect( hwnd, &rc );
				x -= ( rc.right - rc.left )/2;
				y -= ( rc.bottom - rc.top )/2;
				MoveWindow( hwnd, x, y, ( rc.right - rc.left )+1, ( rc.bottom - rc.top )+1, FALSE );
			}
		}

		if ( '.' == RomMistakes[ ROMs[0] ][ ROMs[1] ][0] )
		{
			SetWindowText( hwnd, "ROM Image Information" );
		}

		if ( lparam )
		{
			SetDlgItemText( hwnd, IDOK, "&Close" );
			ShowWindow( GetDlgItem( hwnd, IDCANCEL ), SW_HIDE );
		}
		SetDlgItemText(hwnd, IDC_ROM_ERROR, &RomMistakes[ ROMs[0] ][ ROMs[1] ][1] );
		SetDlgItemText(hwnd, IDC_ROM1_DESC,	RomTypeNames[ ROMs[ 0 ]] );
		SetDlgItemText(hwnd, IDC_ROM2_DESC,	RomTypeNames[ ROMs[ 1 ]] );
		break;
    default:
		return FALSE;
	}
	return TRUE;
}


int WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int)
{
	HANDLE commhandle = NULL;
	UWORD *pRom;
	unsigned cbRomSize;
	
	if (strnicmp(Settings.CommPort, "COM", 3) == 0) 
	{
		commhandle = CreateFile(Settings.CommPort, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
		if (commhandle == INVALID_HANDLE_VALUE) 
		{
			char buf[MAX_PATH];
			wsprintf(buf, "Communications port %s already in use", Settings.CommPort);
			MessageBox(NULL, buf, "Copilot", MB_OK);
			commhandle = NULL;
		}
	}

	if ( !load_rom( "palm.rom", &pRom, &cbRomSize ))
	{
		if ( !load_rom( "pilot.rom", &pRom, &cbRomSize ))
		{
			MessageBox(NULL, "Could not load ROM from either \"pilot.rom\" or \"palm.rom\".", "Copilot", MB_OK );
			return( 0 );
		}
	}

	if ( cbRomSize > 0 )
	{
		ROMs[ 0 ] = Garbage;

		if ( cbRomSize >= 0x080000 )
		{
			ROMs[ 0 ] =GetRomType(( char * )pRom );

			if ( cbRomSize > 0x080000 )
			{
				ROMs[ 1 ] = Garbage;

				if ( cbRomSize >= 0x100000 )
				{
					ROMs[ 1 ] = GetRomType(( char * )&pRom[0x40000]);
				}
			}
		}
	}

	if (! ((( ROMs[ 0 ] == Pilot ) && ( ROMs[ 1 ] == Blank )) ||
		   (( ROMs[ 0 ] == Palm_Bank1 ) && ( ROMs[ 1 ] == Palm_Bank2 ))))
	{
         if ( IDOK == DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ROM_ERROR), NULL, RomDlgProc))
		 {
			 return 0;
		 }
	}

    for (unsigned i = 0; i < cbRomSize/2; i++) 
	{
		UWORD *p = pRom + i;
		unsigned char *bp = (unsigned char *)p;
		*p = (*bp << 8) | *(bp+1);
    }

	int r = CPU_init(commhandle, Settings.RamSize, pRom, cbRomSize);
	if (r != 0) 
	{
		MessageBox(NULL, CpuInitErrors[r], "Copilot", MB_OK);
		return 0;
	}

	WaitForSingleObject(StartLCD(), INFINITE);

	return 0;
}

bool load_rom( char * name, UWORD **prommemory, unsigned *promsize)
{
	UWORD *rommemory;
	unsigned rom_size;

    HANDLE f = CreateFile(name, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

    if (f == INVALID_HANDLE_VALUE) 
	{
		char Filename[_MAX_PATH];

		GetModuleFileNameA( NULL, Filename, sizeof( Filename ));

		char * pchLastSlash = NULL;
		for ( char * pch = Filename; *pch != '\0'; pch ++ )
		{
			if ( *pch == '\\' )
				pchLastSlash = pch;
		}

		// Ensure room to append "pilot.rom"
		if (( NULL != pchLastSlash ) && (( pch - Filename ) < ( sizeof( Filename ) - 11 )))
		{
			strcpy( pchLastSlash + 1, name );
			f = CreateFile(Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);		
		}


	    if (f == INVALID_HANDLE_VALUE) {
		    return false;
		}
	}
    
	rom_size = GetFileSize( f, NULL );
	*promsize = rom_size;

	rommemory = (UWORD *)VirtualAlloc(NULL, rom_size, MEM_COMMIT, PAGE_READWRITE);
	*prommemory = rommemory;

    if (rommemory == NULL) 
	{
        return false;
    }
    
    unsigned long n;
    if (!ReadFile(f, rommemory, rom_size, &n, NULL) || n != rom_size) 
	{
        return false;
    }
    CloseHandle(f);
    
    return true;
}
